#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
@author: Guillaume Laurent
Comparison of Singular Value Decomposition (SVD)
on CPU (NumPy and SciPy) and on NVIDIA GPU (scikit-cuda)
"""

from __future__ import division, print_function
import numpy as np
from scipy import dot as sp_dot
from scipy import linalg as sp_linalg
import time

# GPU libraries
from pycuda import autoinit     # needed
from pycuda import gpuarray as gpuarray
from skcuda import linalg as cu_linalg
cu_linalg.init()                # needed

# Select matrix dimensions and threshold
n = 1024        # number of columns
m = n + 1       # number of rows
k = 25
type_list = [np.float32, np.complex64, np.float64, np.complex128]

# Functions definition
def svd_numpy(X, t, k):
    """
    SIngular Value Decomposition with numpy
    Usage:  svd_numpy(X, t, k)
    Input:  X       matrix to decompose (array)
            t       data type (string)
            k       threshold for low-rank approximation (integer)
    Output: none
    """
    print('\n\n\nNumpy svd for type ' + str(np.dtype(t)))
    t_0 = time.time()
    # Matrix decomposition
    X = X.astype(t)
    U, S, Vh = np.linalg.svd(X)
    t_1 = time.time()
    # Matrix reconstruction
    Smat = np.zeros((m, n)).astype(t)
    Smat[:min(m, n),:min(m, n)] = np.diag(S)
    X_rec = np.dot(U[:,:k], np.dot(Smat[:k,:k], Vh[:k,:]))
    t_2 = time.time()
    print('Decomposition time:\t\t{0:8.2f}s'.format(t_1 - t_0))
    print('Resconstruction time:\t\t{0:8.2f}s'.format(t_2 - t_1))
    print('Maximum relative error:\t\t  {0:10.2e}' \
        .format(np.max(np.abs(X-X_rec))/np.max(np.abs(X))))

def svd_scipy(X, t, k):
    """
    SIngular Value Decomposition with scipy
    Usage:  svd_scipy(X, t, k)
    Input:  X       matrix to decompose (array)
            t       data type (string)
            k       threshold for low-rank approximation (integer)
    Output: none
    """
    print('\nScipy svd for type ' + str(np.dtype(t)))
    t_0 = time.time()
    # Matrix decomposition
    X = X.astype(t)
    U, S, Vh = sp_linalg.svd(X)
    t_1 = time.time()
    # Matrix reconstruction
    Smat = sp_linalg.diagsvd(S, m, n)
    X_rec = sp_dot(U[:,:k], sp_dot(Smat[:k,:k], Vh[:k,:]))
    t_2 = time.time()
    print('Decomposition time:\t\t{0:8.2f}s'.format(t_1 - t_0))
    print('Resconstruction time:\t\t{0:8.2f}s'.format(t_2 - t_1))
    print('Maximum relative error:\t\t  {0:10.2e}' \
        .format(np.max(np.abs(X-X_rec))/np.max(np.abs(X))))

def svd_skcuda(X, t, k):
    """
    SIngular Value Decomposition with scikit-cuda and CULA library
    Usage:  svd_numpy(X, t, k)
    Input:  X       matrix to decompose (array)
            t       data type (string)
                    CULA free only support t = np.float32, np.complex64
            k       threshold for low-rank approximation (integer)
    Output: none
    """
    if t not in [np.float32, np.complex64]:
        print('\n{} not supported with CULA free'.format(t))
        return
    
    print('\nskCuda svd for type ' + str(np.dtype(t)))
    t_0 = time.time()
    # Matrix decomposition
    X = X.astype(t)
    X_gpu = gpuarray.to_gpu(X)  # send data to gpu
    U_gpu, S_gpu, Vh_gpu = cu_linalg.svd(X_gpu, lib='cula')
    X_gpu.gpudata.free()        # release gpu memory
    t_1 = time.time()
    # Matrix reconstruction
    Smat_gpu = gpuarray.zeros((m, n), t)
    Smat_gpu[:min(m, n),:min(m, n)] = cu_linalg.diag(S_gpu).astype(t)
    X_rec_gpu = (cu_linalg.dot(U_gpu[:,:k], \
        cu_linalg.dot(Smat_gpu[:k,:k], Vh_gpu[:k,:])))
    X_rec = X_rec_gpu.get()     # take data from gpu
    X_rec_gpu.gpudata.free()    # release gpu memory
    U_gpu.gpudata.free()
    S_gpu.gpudata.free()
    Smat_gpu.gpudata.free()
    Vh_gpu.gpudata.free()
    t_2 = time.time()
    print('Decomposition time:\t\t{0:8.2f}s'.format(t_1 - t_0))
    print('Resconstruction time:\t\t{0:8.2f}s'.format(t_2 - t_1))
    print('Maximum relative error:\t\t  {0:10.2e}' \
        .format(np.max(np.abs(X-X_rec))/np.max(np.abs(X))))

# Main Program
for t in type_list:
    X = np.arange(m*n).reshape(m, n)
    svd_numpy(X, t, k)
    svd_scipy(X, t, k)
    svd_skcuda(X, t, k)
